from os import environ as envvars
from dataclasses import asdict
from textwrap import dedent
from unittest import TestCase
from unittest.mock import patch

from tests.recruitment.agency import acceptable_broker_names
from tests.recruitment.agency import fake_credentials
from recruitment.agency import Config


class ConfigTest(TestCase):

    broker_name = acceptable_broker_names[0]  # kinesis
    expected_serialization = dedent(
        f"""\
        service_name={broker_name}
        region_name=somewhere-in-the-world
        access_key_id=s3curityBadge!
        secret_access_key=p@ssw0rd!
        endpoint_url=some-computer.com"""
    )
    expected_serialization_missing_values = dedent(
        f"""\
        service_name={broker_name}
        region_name=
        access_key_id=
        secret_access_key=
        endpoint_url="""
    )

    def test_cannot_instantiate_given_invalid_Broker_type(self):
        with self.assertRaises(ValueError):
            Config('invalid Broker type')

    def test_can_instantiate(self):
        for name in acceptable_broker_names:
            config = Config(name, **fake_credentials)
            config_dict = asdict(config)
            config_dict.pop('service_name')
            self.assertDictEqual(config_dict, fake_credentials)

    def test_can_serialize(self):
        config = Config(self.broker_name, **fake_credentials)
        self.assertEqual(str(config), self.expected_serialization)

    def test_can_serialize_given_absent_values(self):
        config = Config(self.broker_name)
        self.assertEqual(str(config), self.expected_serialization_missing_values)

    def test_can_serialize_asfile(self):
        profile_name = 'testing'
        config = Config(self.broker_name, **fake_credentials)
        self.assertEqual(
            config.asfile(profile=profile_name),
            f'[{profile_name}]\n' + self.expected_serialization,
        )
        config_missing_values = Config(self.broker_name)
        self.assertEqual(
            config_missing_values.asfile(profile=profile_name),
            f'[{profile_name}]\n' + self.expected_serialization_missing_values,
        )

    @patch.dict(
        envvars,
        {
            'AWS_SERVICE_NAME': broker_name,
            'AWS_REGION_NAME': fake_credentials['region_name'],
            'AWS_ACCESS_KEY_ID': fake_credentials['access_key_id'],
            'AWS_SECRET_ACCESS_KEY': fake_credentials['secret_access_key'],
            'AWS_ENDPOINT_URL': fake_credentials['endpoint_url'],
        },
        clear=True
    )
    def test_can_constitute_from_environment_variabiles(self):
        config = Config.fromenv()
        self.assertEqual(config.region_name, fake_credentials['region_name'])
        self.assertEqual(config.access_key_id, fake_credentials['access_key_id'])
        self.assertEqual(config.secret_access_key, fake_credentials['secret_access_key'])
        self.assertEqual(config.endpoint_url, fake_credentials['endpoint_url'])

    @patch.dict(
        envvars,
        {
            'AWS_SERVICE_NAME': broker_name,
            'AWS_REGION_NAME': fake_credentials['region_name'],
            'AWS_ACCESS_KEY_ID': fake_credentials['access_key_id'],
            'AWS_SECRET_ACCESS_KEY': fake_credentials['secret_access_key'],
            'AWS_ENDPOINT_URL': fake_credentials['endpoint_url'],
        },
        clear=True
    )
    def test_can_supplement_Config_instance_from_environment(self):
        some_service_name = 'logs'
        self.assertNotEqual(self.broker_name, some_service_name)
        config = Config(service_name=some_service_name)
        self.assertEqual(config.service_name, some_service_name)
        self.assertEqual(config.region_name, None)
        self.assertEqual(config.access_key_id, None)
        self.assertEqual(config.secret_access_key, None)
        self.assertEqual(config.endpoint_url, None)
        config = config.supplement('env')
        self.assertEqual(config.service_name, some_service_name)
        self.assertEqual(config.region_name, fake_credentials['region_name'])
        self.assertEqual(config.access_key_id, fake_credentials['access_key_id'])
        self.assertEqual(config.secret_access_key, fake_credentials['secret_access_key'])
        self.assertEqual(config.endpoint_url, fake_credentials['endpoint_url'])

    def test_cannot_instantiate_without_service_name(self):
        with self.assertRaises(Config.AttributeDeclaredIncorrectly):
            Config(None)

    def test_cannot_instantiate_with_invalid_service_name(self):
        with self.assertRaises(Config.AttributeDeclaredIncorrectly):
            Config(1)



from textwrap import dedent
from unittest import TestCase
from unittest.mock import ANY
from unittest.mock import patch

from botocore.exceptions import NoRegionError
from botocore.stub import Stubber

from tests.recruitment.agency import fake_credentials
from tests.recruitment.agency import client
from tests.recruitment.agency import raise_this
from recruitment.agency import Broker
from recruitment.agency import Config
from recruitment.agency import Communicator
from recruitment.agency.temp import Communicator as FakeCommunicator


class CommunicatorTest(TestCase):

    @patch('boto3.client')
    def test_cannot_instantiate_with_invalid_Config(self, mock_boto_client):
        mock_boto_client.side_effect = raise_this(exception=ValueError)
        with self.assertRaises(Communicator.FailedToInstantiate):
            Communicator(config=Config(Broker.sns))

        mock_boto_client.side_effect = raise_this(exception=NoRegionError)
        with self.assertRaises(Communicator.FailedToInstantiate):
            Communicator(config=Config(Broker.sns))

    @patch('boto3.client')
    def test_communicator_instantiation_failure_redacts_secrets(self, mock_boto_client):
        expected_serialization = dedent(
            """\
            service_name=sns
            region_name=
            access_key_id=**********
            secret_access_key=**********
            endpoint_url="""
        )

        mock_boto_client.side_effect = raise_this(exception=ValueError)
        with self.assertRaises(Communicator.FailedToInstantiate) as error_ctx:
            Communicator(config=Config(Broker.sns))

        self.assertEqual(str(error_ctx.exception), expected_serialization)

    @patch('boto3.client')
    def test_can_create_topic_and_send_message(self, mock_boto_client):
        region = 'some-region-1'
        topic_name = 'some-topic'
        broker = Broker.sns

        expected_topic_response = {
            'TopicArn': f'arn:aws:sns:{region}:12345:{topic_name}'
        }
        expected_publish_response = {
            'MessageId': '00000000-0000-0000-0000-000000000000'
        }

        sns = client(broker.name, region)
        mock_boto_client.return_value = sns
        with Stubber(sns) as stubber:
            stubber.add_response(
                broker.interface['declare_receiver'], expected_topic_response
            )
            stubber.add_response(broker.interface['send'], expected_publish_response)

            commlink = Communicator(Config(broker, **fake_credentials))
            topic_creation = commlink.declare_receiver(
                Name='string',
                Attributes={'string': 'string'},
                Tags=[
                    {'Key': 'string', 'Value': 'string'},
                ],
            )
            message_receipt = commlink.send(Message='some message!')

        self.assertEqual(topic_creation, expected_topic_response)
        self.assertEqual(message_receipt, expected_publish_response)

    @patch('boto3.client')
    def test_can_receive_logs_message(self, mock_boto_client):
        region = 'some-region-1'
        broker = Broker.logs

        expected_response = {
            'events': [
                    {
                    'timestamp': 1663600455651,
                    'message': '  File "/usr/local/lib/python3.7/http/client.py", line 1026, in _send_output',
                    'ingestionTime': 1663600458020
                }
            ],
            'nextForwardToken': '-->',
            'nextBackwardToken': '<--',
            'ResponseMetadata': {
                'RequestId': 'b14a4a55-5ab1-4da0-aa90-eafaa721fdaa',
                'HTTPStatusCode': 200,
                'HTTPHeaders': {
                    'x-amzn-requestid': 'b14a4a55-5ab1-4da0-aa90-eafaa721fdaa',
                    'content-type': 'application/x-amz-json-1.1',
                    'content-length': '1057709',
                    'date': 'Mon, 19 Sep 2022 15:59:05 GMT'
                },
                'RetryAttempts': 0
            }
        }

        logs = client(broker.name, region)
        mock_boto_client.return_value = logs
        with Stubber(logs) as stubber:
            stubber.add_response(
                broker.interface['receive'], expected_response
            )

            commlink = Communicator(Config(broker, **fake_credentials))
            message_receipt = commlink.receive(logGroupName='someLogGroupName', logStreamName='someLogStreamName')

        self.assertEqual(message_receipt, expected_response)


class TempCommunicatorTest(TestCase):

    def test_context(self):
        config = Config('logs', 'us-east-1','some-acces-key-id', 'some-secret', 'https://some-endpoint.com')
        payload_items = [('testing', 123)]
        logGroupName, logStreamName = 'testing', '123'
        expectations = {
            'expected_payload': dict(payload_items),
            'expected_args': (
                ANY,
                {
                    'logGroupName': logGroupName,
                    'logStreamName': logStreamName
                }
            )
        }
        with FakeCommunicator(config, **expectations) as fake_commlink:
            received = fake_commlink.receive(logGroupName=logGroupName, logStreamName=logStreamName)

        self.assertCountEqual(payload_items, received.items())



from typing import Callable
from typing import Optional
from typing import Protocol
from typing import runtime_checkable

from actionpack import Action
from actionpack.actions import RetryPolicy
from actionpack.actions import Write


@runtime_checkable
class HasContingency(Protocol):
    retry_policy_provider: Optional[Callable[[Action], RetryPolicy]] = None
    record_failure_provider: Optional[Callable[[], Write]] = None



from unittest import TestCase
from unittest.mock import patch

from recruitment.agency import Agent
from recruitment.agency import Config
from recruitment.agency import Consumer
from recruitment.agency import Publisher
from recruitment.agency.protocols import HasContingency

from tests.recruitment.agency import acceptable_broker_names


class HasContingencyTest(TestCase):

    @patch('boto3.client')
    def test_fully_implemented_entities_follow_protol(self, mock_boto_client):
        broker_name = acceptable_broker_names[0]
        config = Config(broker_name)
        self.assertIsInstance(Agent(config), HasContingency)
        self.assertIsInstance(Publisher(config), HasContingency)

    @patch('boto3.client')
    def test_partial_entity_violates_protocol(self, mock_boto_client):
        self.assertNotIsInstance(Consumer(), HasContingency)



from enum import auto
from enum import Enum
from typing import Dict
from typing import Optional


class NaturalEnum(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name


class Broker(NaturalEnum):
    """A repository for declaring services and their interfaces"""

    logs = auto()
    s3 = auto()
    sns = auto()
    sqs = auto()
    kinesis = auto()

    @property
    def interface(self) -> Dict[str, Optional[str]]:
        send = 'send'
        declare_receiver = 'declare_receiver'
        receive = 'receive'
        methods_for = {
            Broker.logs: {receive: 'get_log_events'},
            Broker.s3: {send: 'upload_fileobj', declare_receiver: 'create_bucket'},
            Broker.sns: {send: 'publish', declare_receiver: 'create_topic'},
            Broker.sqs: {send: 'send_message', declare_receiver: 'create_queue'},
            Broker.kinesis: {send: 'put_record', declare_receiver: 'create_stream'},
        }
        return methods_for[self]


class From(NaturalEnum):
    env = auto()
    file = auto()


class CloudProvider(NaturalEnum):
    AWS = auto()
    


from typing import Iterable
from typing import Optional
from unittest.mock import patch

from recruitment.agency import Communicator as ActualCommunicator
from recruitment.agency import Config


class Communicator(ActualCommunicator):

    def __init__(
        self,
        config: Config,
        expected_payload: Optional[dict] = None,
        expected_args: Optional[Iterable] = None,
        expected_kwargs: Optional[dict] = None
    ):
        self.response_provider = lambda: expected_payload or {}
        self.args_provider = lambda: expected_args or {}
        self.kwargs_provider = lambda: expected_kwargs or {}
        super().__init__(config)

    def __enter__(self):
        self.patcher = patch('botocore.client.BaseClient._make_api_call')
        self.mock = self.patcher.start()
        self.mock.return_value = self.response_provider()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Given a call signature for `botocore.client.BaseClient._make_api_call` as follows:
            def _make_api_call(self, operation_name, api_params):
                ...
        where, for example, operation_name and api_params could be:
            * GetLogEvents
            * {'logGroupName': ..., 'logStreamName': ...}
        """
        self.mock.assert_called_with(*self.args_provider(), **self.kwargs_provider())
        self.patcher.stop()
